@@ -25,6 +25,8 @@
 
 #define MESSAGE_TAGS_DATA_INIT			\
 	{									\
+		NULL, /* account name */		\
+		FALSE, /* identified to nick */ \
 		(time_t)0, /* timestamp */		\
 	}
 
@@ -36,9 +38,13 @@
  */
 typedef struct 
 {
+	char *account;
+	gboolean identified;
 	time_t timestamp;
 } message_tags_data;
 
+void message_tags_data_free (message_tags_data *tags_data);
+
 void proto_fill_her_up (server *serv);
 
 #endif
